///////////////////////////////////////////////////////////////////////////////
//
// Time-stamp: "09.08.09 19:57 AVR_TSic206_TSic306.cpp klaus?wachtler.de"
//
////////////////////////////////////////////////////////////////////////////////
//
// Beispiel fr Anschlu Temperatursensor TSic 206/306 (ZACwire) an ATMega32
//
// Zudem verwendet:
// - 44780 (mind. 2x16)
// - 2 low current-LEDs (2 mA) mit 820 Ohm an PC6 und PC7
//   Die an PC7 leuchtet, solange eine Messung luft; die an PC6
//   leuchtet, solange der AVR mit der Messung wirklich beschftigt
//   ist (busy).
//
//
/////////////////////////////////////////////////////////////////////////////
//
//    ATMega32:
//
//                               +----_----+
//  LCD-DB4 ------  (XCK/T0) PB0 |  1   40 | PA0 (ADC0)
//  LCD-DB5 ------      (T1) PB1 |  2   39 | PA1 (ADC1)
//  LCD-DB6 ---- (INT2/AIN0) PB2 |  3   38 | PA2 (ADC2)
//  LCD-DB7 ----- (OC0/AIN1) PB3 |  4   37 | PA3 (ADC3)
//  LCD-RS ----------- (-SS) PB4 |  5   36 | PA4 (ADC4)
//  LCD-EN ---------- (MOSI) PB5 |  6   35 | PA5 (ADC5)
//                    (MISO) PB6 |  7   34 | PA6 (ADC6)
//                     (SCK) PB7 |  8   33 | PA7 (ADC7)
//                        -RESET |  9   32 | AREF
//                           VCC | 10   31 | GND
//                           GND | 11   30 | AVCC
//                         XTAL2 | 12   29 | PC7 (TOSC2) -- LED2-R820-GND
//                         XTAL1 | 13   28 | PC6)(TOSC1) -- LED1-R820-GND
//                     (RxD) PD0 | 14   27 | PC5)(TDI)
//                     (TxD) PD1 | 15   26 | PC4 (TDO)
//  TSic 2 Data ----  (INT0) PD2 | 16   25 | PC3 (TMS)
//                    (INT1) PD3 | 17   24 | PC2 (TCK)
//                    (OC1B) PD4 | 18   23 | PC1 (SDA)
//                    (OC1A) PD5 | 19   22 | PC0 (SCL)
//  TSic 1 VCC ------ (ICP1) PD6 | 20   21 | PD7 (OC2)
//                               +---------+
//
//  TSic 3 GND ---- GND
//
// Sowie die restlichen LCD-Leitungen:
//   VSS  1-----GND
//   VDD  2-----+5V
//   VEE  3--Abgriff Poti 10k zw. GND und +5V
//   R/W  5-----GND
//   DB0  7-NC
//   DB1  8-NC
//   DB2  9-NC
//   DB3 10-NC
// LED-A 15--R 220 Ohm--+5V;  Achtung! R je nach LCD!
// LED-K 15-----GND
//
////////////////////////////////////////////////////////////////////////////////



//#include <stdlib.h>
#include <stdio.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <inttypes.h>
#include <util/delay.h>

#include "LCD44780.h"


// Port und Pinnummer fr LEDs:
#define PORT_LED_RUNNING       PORTC
#define PINNUMMER_LED_RUNNING  (7)
#define PORT_LED_BUSY          PORTC
#define PINNUMMER_LED_BUSY     (6)

#define DDR_LED_RUNNING        (((uint8_t*)&PORT_LED_RUNNING)[-1])
#define DDR_LED_BUSY           (((uint8_t*)&PORT_LED_BUSY)[-1])


#define LED_RUNNING_ON  (PORT_LED_RUNNING|=(1<<PINNUMMER_LED_RUNNING))
#define LED_RUNNING_OFF (PORT_LED_RUNNING&=~(1<<PINNUMMER_LED_RUNNING))

#define LED_BUSY_ON     (PORT_LED_BUSY|=(1<<PINNUMMER_LED_BUSY))
#define LED_BUSY_OFF    (PORT_LED_BUSY&=~(1<<PINNUMMER_LED_BUSY))




/////////////////////////////////////////////////////////////////////////////

// Auslesen eines Temperaturfhlers TSic 206/306
//
// [3] TSic (tm)
//     Technical Notes - ZACWire (tm) Digital Output
//     ZMD
//
// Das Signal ist normalerweise high und wird fr jedes Bit auf Low
// gezogen.
// Es besteht aus 11 Nutzbits; mit dem ZACWire-Protokoll knnen in
// einem Paket 8 Bits bertragen werden. Deshalb werden je Wert 2
// Bytes bertragen, wobei im oberen Byte 5 Nullbytes zum Auffllen
// stehen.
//
// Jedes Bit dauert 125 usec, also hat man eine Bitrate von 8 kHz.
//
// Eine Sequenz besteht aus folgenden Bits:
// [ 0]  Startbit               __--
// [ 1]  upper byte bit 7       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 2]  upper byte bit 6       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 3]  upper byte bit 5       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 4]  upper byte bit 4       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 5]  upper byte bit 3       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 6]  upper byte bit 2       _--- fr 1, ___- fr 0
// [ 7]  upper byte bit 1       _--- fr 1, ___- fr 0
// [ 8]  upper byte bit 0       _--- fr 1, ___- fr 0
// [ 9]  hParity                _--- fr 1, ___- fr 0 (even parity des upper byte)
// [10]  Stopbit                --__
// [11]  Startbit               __--
// [12]  lower byte bit 7       _--- fr 1, ___- fr 0
// [13]  lower byte bit 6       _--- fr 1, ___- fr 0
// [14]  lower byte bit 5       _--- fr 1, ___- fr 0
// [15]  lower byte bit 4       _--- fr 1, ___- fr 0
// [16]  lower byte bit 3       _--- fr 1, ___- fr 0
// [17]  lower byte bit 2       _--- fr 1, ___- fr 0
// [18]  lower byte bit 1       _--- fr 1, ___- fr 0
// [19]  lower byte bit 0       _--- fr 1, ___- fr 0
// [20]  lParity                _--- fr 1, ___- fr 0 (even parity des lower byte)
//
//
//
// Prinzipiell hat man zwei Mglichkeiten, an die Temperatur zu
// kommen (siehe [3]):
// - Entweder der Sensor ist dauernd an, dann liefert er pro Sekunde 10
//   Werte.
// - Oder man schaltet ihn fr eine Messung an, nach etwa 65...85 msec
//   liefert er den ersten Wert, danach schaltet man ihn wieder ab.
//   Der Vorteil dabei ist, da man nicht 10 mal pro Sekunde Rechenzeit
//   verbrt fr die zeitlich enge Abfrage, und zudem das Auslesen zu
//   einem irgendwie gnstigen Zeitpunkt machen kann (z.B. wenn wenig
//   Rechenlast bentigt wird, oder der bentigte Timer gerade nicht
//   anderweitig benutzt wird).
//
// Vorgehensweise:
// Der Sensor wird fr jede Messung eingeschaltet.
// Anschlieend wird INT0 auf fallende Flanke aktiviert.
// In der ISR-Routine dazu wird INT0 gleich wieder deaktivert, ein
// Timer (Timer/Counter0) angeworfen und in dessen ISR dann die
// Auswertung gemacht. Eigentlich msste man hier im Raster von 125/2
// = 62.5 usec auslesen, weil jeweils zur Bitmitte das Signal gelesen
// werden kann (_--- steht in der Mitte auf 1, ___- auf 0).
// Das passt aber schlecht zu 16 oder 8 MHz Frequenz bei den mglichen
// Teilerverhltnissen /8, /64, /256 und /1024.
// Eigentlich mu man ja auch nicht genau in der Bitmitte lesen;
// es reicht im Bereich von 31.25 usec bis 93.75 usec ab Bitanfang den
// Wert zu lesen (zweites oder drittes Viertel des Bits).
// Deshalb knnte man den Prescaler auf /256 setzen, das macht eine
// Periodendauer von 16 usec (bei 16 MHz).
// Beim ersten Durchlauf der Timer-ISR, die mindestens 54 usec nach
// Bitanfang liegt, wird der Wert gelesen (also etwa Bitmitte +/- 8
// usec). Diesen Wert nenne ich hier nTimer0.
// Anderes Beispiel:


// Bit: Mitte usec min.Sample         nTimer0     Bedeutung              Verlauf
//                 @16MHz    @8MHz   @16MHz @8MHz
// [ 0]    62.50    54.50    46.50     4     2    Startbit               __--
// [ 1]   187.50   179.50   171.50    12     6    upper byte bit 7       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 2]   312.50   304.50   296.50    20    10    upper byte bit 6       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 3]   437.50   429.50   421.50    27    14    upper byte bit 5       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 4]   562.50   554.50   546.50    35    18    upper byte bit 4       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 5]   687.50   679.50   671.50    43    21    upper byte bit 3       _--- fr 1, ___- fr 0 (immer ___- fr 0, da Fllbit)
// [ 6]   812.50   804.50   796.50    51    25    upper byte bit 2       _--- fr 1, ___- fr 0
// [ 7]   937.50   929.50   921.50    59    29    upper byte bit 1       _--- fr 1, ___- fr 0
// [ 8]  1062.50  1054.50  1046.50    66    33    upper byte bit 0       _--- fr 1, ___- fr 0
// [ 9]  1187.50  1179.50  1171.50    74    37    hParity                _--- fr 1, ___- fr 0 (even parity des upper byte)
// [10]  1312.50  1304.50  1296.50    82    41    Stopbit                --__
// [11]  1437.50  1429.50  1421.50    90    45    Startbit               __--
// [12]  1562.50  1554.50  1546.50    98    49    lower byte bit 7       _--- fr 1, ___- fr 0
// [13]  1687.50  1679.50  1671.50   105    53    lower byte bit 6       _--- fr 1, ___- fr 0
// [14]  1812.50  1804.50  1796.50   113    57    lower byte bit 5       _--- fr 1, ___- fr 0
// [15]  1937.50  1929.50  1921.50   121    61    lower byte bit 4       _--- fr 1, ___- fr 0
// [16]  2062.50  2054.50  2046.50   129    64    lower byte bit 3       _--- fr 1, ___- fr 0
// [17]  2187.50  2179.50  2171.50   137    68    lower byte bit 2       _--- fr 1, ___- fr 0
// [18]  2312.50  2304.50  2296.50   145    72    lower byte bit 1       _--- fr 1, ___- fr 0
// [19]  2437.50  2429.50  2421.50   152    76    lower byte bit 0       _--- fr 1, ___- fr 0
// [20]  2562.50  2554.50  2546.50   160    80    lParity                _--- fr 1, ___- fr 0 (even parity des lower byte)
//
// In der obigen Tabelle steht in der Spalte "Mitte usec" die Zeit der
// jeweiligen Bitmitte, gezhlt ab der fallenden Flanke des ersten
// Starbits.
// In "min.Sample" steht die Zeit ab der ersten fallenden Flanke, bei der
// frhestens gelesen werden kann, also in dem jeweiligen Bit
// (62.5 usec - halbe Abtastperiode ) verstrichen sind.
// nTimer0 ist die Nummer des ersten Timer0-Durchlaufs, der mindestens
// min.Sample nach Bitanfang liegt und daraufhin das Bit liest. Der Wert
// von nTimer0 gilt nur bei 16 bzw. 8 MHz, fr andere Frequenzen
// entsprechend anders zu berechnen.
// Fr andere CPU-Frequenzen oder prescale-Werte lassen sich die
// passenden Werte fr nTimer0 mit folgendem Programm berechnen:
//   // Time-stamp: "05.08.09 19:00 calc_nTimer0_fuer_TSic.c klaus?wachtler.de"
//   #include <stdio.h>
//   int main( int nargs, char **args )
//   {
//     const double f_cpu = 16000000;
//     const unsigned prescaler = 256;
//     const double periodendauer = 1.0/(f_cpu/prescaler);
//     size_t  i;
//
//     printf( "F_CPU = %G\n", f_cpu );
//     printf( "Periodendauer = %G\n", periodendauer );
//     printf( " Bit# | Bitmitte usec | min.Sample usec | nTimer0\n" );
//
//     for( i=0; i<=20; ++i )
//     {
//       printf( "   %2d | %13.2f |   %13.2f | %6ld\n",
//               (int)i,
//               i*125+62.5,
//               (i*125E-6+(62.5E-6-0.5*periodendauer))*1.E6,
//               (long)((i*125E-6+(62.5E-6-0.5*periodendauer))/periodendauer+1)
//               );
//     }
//
//     return 0;
//   }
//
// Beispielausgabe fr 16MHz und prescale=256:
//   F_CPU = 1.6E+07
//   Periodendauer = 1.6E-05
//    Bit# | Bitmitte usec | min.Sample usec | nTimer0
//       0 |         62.50 |           54.50 |      4
//       1 |        187.50 |          179.50 |     12
//       2 |        312.50 |          304.50 |     20
//       3 |        437.50 |          429.50 |     27
//       4 |        562.50 |          554.50 |     35
//       5 |        687.50 |          679.50 |     43
//       6 |        812.50 |          804.50 |     51
//       7 |        937.50 |          929.50 |     59
//       8 |       1062.50 |         1054.50 |     66
//       9 |       1187.50 |         1179.50 |     74
//      10 |       1312.50 |         1304.50 |     82
//      11 |       1437.50 |         1429.50 |     90
//      12 |       1562.50 |         1554.50 |     98
//      13 |       1687.50 |         1679.50 |    105
//      14 |       1812.50 |         1804.50 |    113
//      15 |       1937.50 |         1929.50 |    121
//      16 |       2062.50 |         2054.50 |    129
//      17 |       2187.50 |         2179.50 |    137
//      18 |       2312.50 |         2304.50 |    145
//      19 |       2437.50 |         2429.50 |    152
//      20 |       2562.50 |         2554.50 |    160
//
// Der prescale-Wert mu so klein gewhlt werden, da im Bereich der
// Bitmitte (also sicher nach 1/4 des Bits und sicher vor 3/4 des
// Bits) mindestens einmal die ISR aufgerufen wird. Die Periodendauer
// mu dazu deutlich kleiner als 62.5 usec sein.
//
/////////////////////////////////////////////////////////////////////////////



// mu zum verwendeten externen Interrupt passen (INT0 ist zumindest
// beim atmega8, atmega16 atmega32 und at90s2313 identisch mit PD2,
// vielleicht bei allen AVR).
// An diesem Pin mu Pin 2 (Signal) des TSic angeschlossen werden,
// erstens um den Interrupt INT0 auszulsen und zweitens anschlieend
// der Wert auszulesen.
#define PORT_TSIC1_DATA      PORTD
#define PINNUMMER_TSIC1_DATA (2)

#define DDR_TSIC1_DATA       (((uint8_t*)&PORT_TSIC1_DATA)[-1])
#define PIN_TSIC1_DATA       (((uint8_t*)&PORT_TSIC1_DATA)[-2])

// Hier mu V+ des TSic angeschlossen werden; mit dem Pin wird der
// Sensor bei Bedarf eingeschaltet.
#define PORT_TSIC1_VCC       PORTD
#define PINNUMMER_TSIC1_VCC  (6)

#define DDR_TSIC1_VCC        (((uint8_t*)&PORT_TSIC1_VCC)[-1])
//#define PIN_TSIC1_VCC        (((uint8_t*)&PORT_TSIC1_VCC)[-2])




volatile uint8_t   bTSic1Running = false;
volatile uint8_t   lbyteTSic1   = 0;
volatile uint8_t   hbyteTSic1   = 0;
volatile uint8_t   lpariTSic1   = 0;
volatile uint8_t   hpariTSic1   = 0;
volatile uint8_t   nTimer0TSic1 = 0;



/* Timer/Counter0 Overflow */
ISR(TIMER0_OVF_vect)
{
  if( bTSic1Running )
  {
    LED_BUSY_ON;
    switch( nTimer0TSic1 )
    {
      // Anm. zu den Werten in den case-labels: siehe Ausgabe des
      // Programms calc_nTimer0_fuer_TSic.c

#if F_CPU==16000000

      case    4: // [ 0]   Startbit
        break;
      case   12: // [ 1]   upper byte bit 7
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<7);
        break;
      case   20: // [ 2]   upper byte bit 6
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<6);
        break;
      case   27: // [ 3]   upper byte bit 5
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<5);
        break;
      case   35: // [ 4]   upper byte bit 4
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<4);
        break;
      case   43: // [ 5]   upper byte bit 3
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<3);
        break;
      case   51: // [ 6]   upper byte bit 2
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<2);
        break;
      case   59: // [ 7]   upper byte bit 1
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<1);
        break;
      case   66: // [ 8]   upper byte bit 0
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<0);
        break;
      case   74: // [ 9]   hParity
        hpariTSic1 = (PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0;
        break;
      case   82: // [10]   Stopbit
        break;
      case   90: // [11]   Startbit
        break;
      case   98: // [12]   lower byte bit 7
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<7);
        break;
      case  105: // [13]   lower byte bit 6
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<6);
        break;
      case  113: // [14]   lower byte bit 5
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<5);
        break;
      case  121: // [15]   lower byte bit 4
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<4);
        break;
      case  129: // [16]   lower byte bit 3
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<3);
        break;
      case  137: // [17]   lower byte bit 2
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<2);
        break;
      case  145: // [18]   lower byte bit 1
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<1);
        break;
      case  152: // [19]   lower byte bit 0
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<0);
        break;
      case  160: // [20]   lParity

#elif F_CPU==8000000

      case    2: // [ 0]   Startbit
        break;
      case   6: // [ 1]   upper byte bit 7
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<7);
        break;
      case   10: // [ 2]   upper byte bit 6
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<6);
        break;
      case   14: // [ 3]   upper byte bit 5
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<5);
        break;
      case   18: // [ 4]   upper byte bit 4
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<4);
        break;
      case   21: // [ 5]   upper byte bit 3
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<3);
        break;
      case   25: // [ 6]   upper byte bit 2
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<2);
        break;
      case   29: // [ 7]   upper byte bit 1
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<1);
        break;
      case   33: // [ 8]   upper byte bit 0
        hbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<0);
        break;
      case   37: // [ 9]   hParity
        hpariTSic1 = (PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0;
        break;
      case   41: // [10]   Stopbit
        break;
      case   45: // [11]   Startbit
        break;
      case   49: // [12]   lower byte bit 7
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<7);
        break;
      case  53: // [13]   lower byte bit 6
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<6);
        break;
      case  57: // [14]   lower byte bit 5
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<5);
        break;
      case  61: // [15]   lower byte bit 4
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<4);
        break;
      case  64: // [16]   lower byte bit 3
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<3);
        break;
      case  68: // [17]   lower byte bit 2
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<2);
        break;
      case  72: // [18]   lower byte bit 1
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<1);
        break;
      case  76: // [19]   lower byte bit 0
        lbyteTSic1 |= (((PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0)<<0);
        break;
      case  80: // [20]   lParity

#else
#error "Erst die Tabelle nTimer0TSic1 fuer aktuelle F_CPU berechnen!"
#endif /* if F_CPU==16000000 */

        // Nach dem letzten Wert kann man den Timer abschalten.
        //
        // Falls mehrere Sensoren angeschlossen sind, darf das
        // natrlich nur geschehen, wenn nicht gerade noch andere
        // Messungen laufen (bTSic2Running etc.)!
        lpariTSic1 = (PIN_TSIC1_DATA & (1<<PINNUMMER_TSIC1_DATA))!=0;

        bTSic1Running = false;
        TIMSK &= ~(1<<TOIE0);  // Timer Overflow Interrupt enable Timer/Counter 0
        TCCR0 &= ~( (1<<CS00)|(1<<CS01)|(1<<CS02) ); // 0b000 -> kein Timer/Counter0

        // weiterhin kann der TSic wieder abgeschaltet werden:
        PORT_TSIC1_VCC &= ~(1<<PINNUMMER_TSIC1_VCC);


        break;


      default :

        break;
    }

    ++nTimer0TSic1;
    LED_BUSY_OFF;
  }

}


/* External Interrupt Request 0 */
ISR(INT0_vect)
{
  if( bTSic1Running )
  {
    LED_BUSY_ON;
    // Fr TSic 206/306-Sensor:
    // Timer 0 starten mit /1
    // Timer/Counter0 starten
    TCCR0 &= ~( (1<<CS00)|(1<<CS01)|(1<<CS02) ); // 0b000 -> kein Timer/Counter0
    TCCR0 |= (1<<CS00); // prescaler /1
    TIMSK |= (1<<TOIE0);

    // sich selbst (INT0) abschalten:
    GICR  &= ~(1<<INT0);  // Interrupt bei INT0 ausschalten

    // Pin auf Eingang stellen:
    DDR_TSIC1_DATA &= ~( 1<PINNUMMER_TSIC1_DATA );

    LED_BUSY_OFF;
  }
}


int main( int nargs, char **args )
{
  LCD44780    lcd1( PORTB, 0, // Port, Pin LCD-DB4
                    PORTB, 1, // Port, Pin LCD-DB5
                    PORTB, 2, // Port, Pin LCD-DB6
                    PORTB, 3, // Port, Pin LCD-DB7
                    PORTB, 4, // Port, Pin LCD-RS
                    PORTB, 5  // Port, Pin LCD-Enable
                    );
  char          tmp_puffer[30];

#ifdef DDR_LED_RUNNING
  DDR_LED_RUNNING |= (1<<PINNUMMER_LED_RUNNING);
#endif
#ifdef DDR_LED_BUSY
  DDR_LED_BUSY |= (1<<PINNUMMER_LED_BUSY);
#endif

  sei();
  GICR  &= ~(1<<INT0);  // Interrupt bei INT0 ausschalten
  TIMSK &= ~(1<<TOIE0);  // Timer Overflow Interrupt enable Timer/Counter 0
  TCCR0 &= ~( (1<<CS00)|(1<<CS01)|(1<<CS02) ); // 0b000 -> kein Timer/Counter0
  // TCCR0 |= (1<<CS00)|(1<<CS02); // /1024
  // TCCR0 |= (1<<CS02); // /256
  // TCCR0 |= (1<<CS00)|(1<<CS01); // /64
  // TCCR0 |= (1<<CS01); // /8
  // TCCR0 |= (1<<CS00); // /1
  // TCCR0 |= (1<<CS01)|(1<<CS02); // on T0 falling
  // TCCR0 |= (1<<CS00)|(1<<CS01)|(1<<CS02); // on T0 rising
  // TIMSK |= (1<<TOIE0);

  // Ausgang, um den Strom fr den TSic schalten zu knnen:
  DDR_TSIC1_VCC |= (1<<PINNUMMER_TSIC1_VCC); // Ausgang


  while( "warten aufs jngste Gericht" )
  {

    // Messung starten:
    LED_RUNNING_ON;

    lbyteTSic1   = 0;
    hbyteTSic1   = 0;
    lpariTSic1   = 0;
    hpariTSic1   = 0;
    nTimer0TSic1 = 0;

    // Strom des TSic einschalten:
    PORT_TSIC1_VCC |= (1<<PINNUMMER_TSIC1_VCC); // Ein
    //_delay_ms( 10 );   // kurz Zeit geben? Scheint entbehrlich zu sein
    bTSic1Running = true;
    MCUCR &= ~( (1<<ISC00)|(1<<ISC01) );
    MCUCR |=  ( (1<<ISC01) ); // Interrupt bei falling edge on INT0
    GICR  |= (1<<INT0);  // Interrupt bei INT0 einschalten

    while( bTSic1Running )
    {
      // drapp drapp drapp ...
      // Ende der Messung abwarten;
      // in der Zwischenzeit knnte man auch etwas sinnvolles tun.
    }

    lcd1.setCursor( 0, 0 );

    // Wert begutachten:

    // Parity hbyteTSic1 prfen:
    if( ( ( (hbyteTSic1&0b10000000)!=0 )
          +
          ( (hbyteTSic1&0b01000000)!=0 )
          +
          ( (hbyteTSic1&0b00100000)!=0 )
          +
          ( (hbyteTSic1&0b00010000)!=0 )
          +
          ( (hbyteTSic1&0b00001000)!=0 )
          +
          ( (hbyteTSic1&0b00000100)!=0 )
          +
          ( (hbyteTSic1&0b00000010)!=0 )
          +
          ( (hbyteTSic1&0b00000001)!=0 )
          )&1
        !=hpariTSic1
        )
    {
      lcd1.writeString( "Parity-Error Hi" );
    }
    // Parity lbyteTSic1 prfen:
    else if( ( ( (lbyteTSic1&0b10000000)!=0 )
               +
               ( (lbyteTSic1&0b01000000)!=0 )
               +
               ( (lbyteTSic1&0b00100000)!=0 )
               +
               ( (lbyteTSic1&0b00010000)!=0 )
               +
               ( (lbyteTSic1&0b00001000)!=0 )
               +
               ( (lbyteTSic1&0b00000100)!=0 )
               +
               ( (lbyteTSic1&0b00000010)!=0 )
               +
               ( (lbyteTSic1&0b00000001)!=0 )
               )&1
             !=lpariTSic1
        )
    {
      lcd1.writeString( "Parity-Error Lo" );
    }
    else if( hbyteTSic1>0x07 )
    {
      lcd1.writeString( "range error    " );
    }
    else
    {
      // aus den beiden Byte die Temperatur in 0.1 Grad C berechnen:
      int wertTSic1 = ( ((uint16_t)hbyteTSic1<<8) | lbyteTSic1 ) - 500;

      lcd1.writeData( 'T' );
      lcd1.writeData( '=' );
      lcd1.writeData( '0'+(wertTSic1/100)%10 );
      lcd1.writeData( '0'+(wertTSic1/10)%10 );
      lcd1.writeData( '.' );
      lcd1.writeData( '0'+(wertTSic1/1)%10 );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );
      lcd1.writeData( ' ' );

    }

    // zur Kontrolle in der zweiten Zeile die nackten Bytes mit Parity
    // ausgeben:
    lcd1.setCursor( 0, 1 );
    sprintf( tmp_puffer, "x%04X_%u:x%04X_%u",
             (unsigned int)hbyteTSic1,
             (unsigned int)hpariTSic1,
             (unsigned int)lbyteTSic1,
             (unsigned int)lpariTSic1
             );
    lcd1.writeString( tmp_puffer );

    LED_RUNNING_OFF;

    _delay_ms( 200 );
  }

  return 0;
} /* main( int nargs, char **args ) */

