// Time-stamp: "05.08.09 20:00 LCD44780.h klaus?wachtler.de"
//
// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
// [1] http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial
// [2] http://www.mikrocontroller.net/topic/88543#751982
//
// nderungen:
//
// 23.06.2009 kw    - von auen nicht aufzurufende Funktionen static
//                    gemacht:
//                    lcd_data(), lcd_command(), lcd_enable()
//                  - gemeinsamen Code aus lcd_command() und lcd_data()
//                    in neue Funktion lcd_byte() ausgelagert
//                  - Shiften und Maskieren in lcd_byte() eliminiert
//                  - in lcd_set_cursor() Adressberechnung korrigiert
//                    fr blaues LCD "Blue Line" von Electronic Assembly
//                    EA W204B-NLW bzw. Reichelt LCD204BBL
//
// 02.08.2009 kw    - auf C++ umgestellt
//                  - Ports und Pinnummern beim Initialisieren
//                    anzugeben (anstatt fest verdrahtet)
//                  - mehrere LCDs nutzbar (nur Enable mu auf
//                    unterschiedlichen Pins liegen; fr ein Display
//                    braucht man also 4 Datenpins, RS und Enable; fr
//                    jedes weitere LCD ein weiteres Enable).
//                  - Zeilen zhlen jetzt ebenso wie Spalten ab 0
//
// Getestet mit
// - LCD:
//   - Electronic Assembly EA W204B-NLW (4x20)
//   - 2x16 aus Pollin Addon Board zu Atmel Evaluation Board
//   - 2x20?
// - Compiler: avr-g++ 4.3.2 (Debian Linux)
//
// TODO: 4x40 mit 2 Controllern

#ifndef __LCD44780_h__
#define __LCD44780_h__

#include <inttypes.h>


class LCD44780
{
 public:

  LCD44780( volatile uint8_t &port_d4, uint8_t pin_d4,
            volatile uint8_t &port_d5, uint8_t pin_d5,
            volatile uint8_t &port_d6, uint8_t pin_d6,
            volatile uint8_t &port_d7, uint8_t pin_d7,
            volatile uint8_t &port_rs, uint8_t pin_rs,
            volatile uint8_t &port_e1, uint8_t pin_e1//,
            //volatile uint8_t &port_e2 = *(uint8_t*)0, uint8_t pin_e2 = (uint8_t)-1   // TODO: enable2
            )
    : port_e1( port_e1 ),
      pin_e1( pin_e1 ),
      //port_e2( port_e2 ),
      //pin_e2( pin_e2 ),
      port_rs( port_rs ),
      pin_rs( pin_rs ),
      port_d4( port_d4 ),
      pin_d4( pin_d4 ),
      port_d5( port_d5 ),
      pin_d5( pin_d5 ),
      port_d6( port_d6 ),
      pin_d6( pin_d6 ),
      port_d7( port_d7 ),
      pin_d7( pin_d7 )
  {
    // Ports auf Ausgang schalten
    // Anm.: die DDR fr alle Ports sind jeweils mit einer um 1
    // niedrigeren Adresse eingeblendet als die PORT...; solange Atmel
    // das nicht ndert, kann man ber (&port_...)[-1] auf das
    // DDR-Register zugreifen.
    (&port_d4)[-1] |= (1<<pin_d4);
    (&port_d5)[-1] |= (1<<pin_d5);
    (&port_d6)[-1] |= (1<<pin_d6);
    (&port_d7)[-1] |= (1<<pin_d7);
    (&port_e1)[-1] |= (1<<pin_e1);
    (&port_rs)[-1] |= (1<<pin_rs);

    // muss 3 mal hintereinander gesendet werden zur Initialisierung

    _delay_ms( 15 );

    port_d4 |=  (1<<pin_d4);
    port_d5 |=  (1<<pin_d5);
    port_d6 &= ~(1<<pin_d6);
    port_d7 &= ~(1<<pin_d7);

    port_rs &= ~(1<<pin_rs);      // RS auf 0
    enable();

    _delay_ms( 5 );
    enable();

    _delay_ms( 1 );
    enable();
    _delay_ms( 1 );

    // 4 Bit Modus aktivieren
    port_d4 &=  ~(1<<pin_d4);
    port_d5 |=   (1<<pin_d5);
    port_d6 &=  ~(1<<pin_d6);
    port_d7 &=  ~(1<<pin_d7);

    enable();
    _delay_ms( 1 );

    // 4Bit / 2 Zeilen / 5x7
    writeCommand( 0x28 );

    // Display ein / Cursor aus / kein Blinken
    writeCommand( 0x0C );

    // inkrement / kein Scrollen
    writeCommand( 0x06 );

    clear();
  }


  // Sendet den Befehl zur Lschung des Displays
  void clear() const
  {
    writeCommand( LCD_CLEAR_DISPLAY );
    _delay_ms( 5 );
  }


  // Sendet den Befehl: Cursor Home
  void home() const
  {
    writeCommand( LCD_CURSOR_HOME);
    _delay_ms( 5 );
  }


  // setzt den Cursor in Zeile y (0..3) Spalte x (0..19)
  void setCursor( uint8_t x, uint8_t y ) const
  {
    uint8_t tmp;

    switch ( y )
    {
      // Originalversion von [1]:
      //case 0: tmp=0x80+0x00+x;  break;    // 1. Zeile
      //case 1: tmp=0x80+0x40+x;  break;    // 2. Zeile
      //case 2: tmp=0x80+0x10+x;  break;    // 3. Zeile
      //case 3: tmp=0x80+0x50+x;  break;    // 4. Zeile

      // Klappt mit 4x20 blau ("Blue Line" von Electronic Assembly
      // EA W204B-NLW bzw. Reichelt LCD204BBL)
      case 0: tmp=0x80+0x00+x;  break;    // 1. Zeile
      case 1: tmp=0x80+0x40+x;  break;    // 2. Zeile
      case 2: tmp=0x80+0x14+x;  break;    // 3. Zeile
      case 3: tmp=0x80+0x54+x;  break;    // 4. Zeile
      default: return;
    }
    writeCommand( tmp );
  }


  // Schreibt einen String auf das LCD
  void writeString( const char *data ) const
  {
    while(*data)
    {
      writeData( *data );
      data++;
    }
  }


  // sendet ein Datenbyte an das LCD
  void writeData( unsigned char data) const
  {
    port_rs |= (1<<pin_rs);        // RS auf 1 setzen
    writeByte( data );
  }


private:

  // erzeugt den Enable-Puls
  void enable() const
  {
    // Bei Problemen ggf. Pause gem Datenblatt des LCD Controllers einfgen

    // http://www.mikrocontroller.net/topic/81974#685882
    port_e1 |= (1<<pin_e1);

    _delay_us( 1 );                   // kurze Pause
    // Bei Problemen ggf. Pause gem Datenblatt des LCD Controllers verlngern
    // http://www.mikrocontroller.net/topic/80900
    port_e1 &= ~(1<<pin_e1);
  }


  // sendet ein Daten- oder Kommandobyte (RS am LCD mu vorher
  // entsprechend gesetzt sein)
  void writeByte( unsigned char data ) const
  {
    // oberes Nibble auf den LCD-Pins 4..7 ausgeben:

    if( data & 0x10 )
    {
      port_d4 |= (1<<pin_d4);
    }
    else
    {
      port_d4 &= ~(1<<pin_d4);
    }
    if( data & 0x20 )
    {
      port_d5 |= (1<<pin_d5);
    }
    else
    {
      port_d5 &= ~(1<<pin_d5);
    }
    if( data & 0x40 )
    {
      port_d6 |= (1<<pin_d6);
    }
    else
    {
      port_d6 &= ~(1<<pin_d6);
    }
    if( data & 0x80 )
    {
      port_d7 |= (1<<pin_d7);
    }
    else
    {
      port_d7 &= ~(1<<pin_d7);
    }
    enable();

    // unteres Nibble auf den LCD-Pins 4..7 ausgeben:

    if( data & 0x01 )
    {
      port_d4 |= (1<<pin_d4);// setzen
    }
    else
    {
      port_d4 &= ~(1<<pin_d4);
    }
    if( data & 0x02 )
    {
      port_d5 |= (1<<pin_d5);
    }
    else
    {
      port_d5 &= ~(1<<pin_d5);
    }
    if( data & 0x04 )
    {
      port_d6 |= (1<<pin_d6);
    }
    else
    {
      port_d6 &= ~(1<<pin_d6);
    }
    if( data & 0x08 )
    {
      port_d7 |= (1<<pin_d7);
    }
    else
    {
      port_d7 &= ~(1<<pin_d7);
    }
    enable();

    // Dem LCD Zeit geben, das Kommando auszufhren
    _delay_us(42);
  }


  // sendet einen Befehl an das LCD
  void writeCommand( unsigned char command ) const
  {
    port_rs &= ~(1<<pin_rs);        // RS auf 0 setzen
    writeByte( command );
  }


  enum
    {
      LCD_CLEAR_DISPLAY = 0x01,
      LCD_CURSOR_HOME   = 0x02,
    };


  volatile uint8_t &port_e1;
  uint8_t           pin_e1;
  //volatile uint8_t &port_e2;
  //uint8_t           pin_e2;
  volatile uint8_t &port_rs;
  uint8_t           pin_rs;
  volatile uint8_t &port_d4;
  uint8_t           pin_d4;
  volatile uint8_t &port_d5;
  uint8_t           pin_d5;
  volatile uint8_t &port_d6;
  uint8_t           pin_d6;
  volatile uint8_t &port_d7;
  uint8_t           pin_d7;

};

#endif /* ifndef __LCD44780_h__ */

// Local Variables:
// mode:C++
// End:
